/* Game engine routines..
TODO:   fix enemy damage
		Player hurtable
		dust clouds
		dash-flash
		player death animation

*/

#include <allegro.h>
#include <fps2d.h>
#include <stdio.h>
#include "main.h"
#include "engine.h"
#include "enginevars.h"
#include "debug.h"

#include "sound.h"
#include "sprites.h"
#include "megamanx.h"

#define ROLLERGUY 0

TBlocks* current_map;
MAP maps[MAXMAPS];
DATAFILE* megamanx;
char exit_level;
PLAYER player;
ENEMYDAT enemies[100];

char temps[32];
char jumppressed;
char shootpressed;
char dashpressed;
char dashsoundplaying;
char total_enemies;

void init_enemies();
void doenemyai();
void rollerguyai( ENEMYDAT &e );

void init_maps();
void draw_screen();
void getinput();
void getinput();
void gametick();
void set_camera();
void draw_player();
void setup_player();
void build_animations();
void update_animation();
void draw_bullets();
BITMAP* sil( BITMAP* bmp, int c );
void draw_trans_sprite_h_flip( BITMAP* dest, BITMAP* bmp, int x, int y );
void draw_lifebar( int x, int y );
void draw_lifemeter( int max, int curr, int x, int y, int size );
void draw_enemies();

void do_playerdeath();

char inair();
char touchingwall();
char stuck();
char einair(ENEMYDAT &e);
char estuck(ENEMYDAT &e);
char reversedir( char dir );

ANIM walking;
ANIM standing;
ANIM jumping;
ANIM falling;
ANIM sliding;
ANIM dashing;
ANIM kneeling;
ANIM hurting;

ANIM swalking;
ANIM sstanding;
ANIM sjumping;
ANIM sfalling;
ANIM sdashing;
ANIM ssliding;
ANIM skneeling;

ANIM bullet0,bullet1,bullet2;

//////////////////////////////////////////////////////////////////////////
///////////////  build_animations() //////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void build_animations()
{
	walking.frames[0]=walking00;
	walking.frames[1]=walking01;
	walking.frames[2]=walking02;
	walking.frames[3]=walking03;
	walking.frames[4]=walking04;
	walking.frames[5]=walking05;
	walking.frames[6]=walking06;
	walking.frames[7]=walking07;
	walking.frames[8]=walking08;
	walking.frames[9]=walking09;
	walking.frames[10]=walking10;
	walking.loopframe=1;
    walking.lastframe=10;
    walking.current_frame=0;

    hurting.frames[0]=hurt00;
	hurting.frames[1]=hurt01;
	hurting.frames[2]=hurt02;
	hurting.frames[3]=hurt03;
	hurting.frames[4]=hurt04;
	hurting.frames[5]=hurt05;
	hurting.frames[6]=hurt06;
	hurting.frames[7]=hurt07;
	hurting.frames[8]=hurt08;
	hurting.frames[9]=hurt09;
	hurting.frames[10]=hurt10;
	hurting.loopframe=0;
	hurting.lastframe=10;
    hurting.current_frame=0;

    swalking.frames[0]=walking00;
	swalking.frames[1]=swalking01;
	swalking.frames[2]=swalking02;
	swalking.frames[3]=swalking03;
	swalking.frames[4]=swalking04;
	swalking.frames[5]=swalking05;
	swalking.frames[6]=swalking06;
	swalking.frames[7]=swalking07;
	swalking.frames[8]=swalking08;
	swalking.frames[9]=swalking09;
	swalking.frames[10]=swalking10;

    jumping.frames[0]=jumping00;
    jumping.frames[1]=jumping01;
    jumping.frames[2]=jumping02;
    jumping.loopframe=2;
    jumping.lastframe=2;
    jumping.current_frame=0;

    sjumping.frames[0]=sjumping00;
	sjumping.frames[1]=sjumping01;
    sjumping.frames[2]=sjumping02;

    sliding.frames[0]=sliding00;
    sliding.frames[1]=sliding01;
    sliding.frames[2]=sliding02;
    sliding.loopframe=2;
    sliding.lastframe=2;
    sliding.current_frame=0;

    ssliding.frames[0]=sliding00;
	ssliding.frames[1]=sliding01;
    ssliding.frames[2]=ssliding00;

    falling.frames[0]=falling00;
	falling.frames[1]=falling01;
	falling.loopframe=1;
	falling.lastframe=1;
    falling.current_frame=0;

	sfalling.frames[0]=sfalling00;
	sfalling.frames[1]=sfalling01;

    dashing.frames[0]=dashing00;
	dashing.frames[1]=dashing01;
	dashing.loopframe=1;
	dashing.lastframe=1;
    dashing.current_frame=0;

	sdashing.frames[0]=sdashing00;
	sdashing.frames[1]=sdashing01;

    kneeling.frames[0]=kneeling00;
	kneeling.loopframe=0;
	kneeling.lastframe=0;
    kneeling.current_frame=0;

    skneeling.frames[0]=skneeling00;

    standing.frames[0]=standing00;
	standing.frames[1]=standing00;
	standing.frames[2]=standing00;
	standing.frames[3]=standing00;
	standing.frames[4]=standing00;
	standing.frames[5]=standing00;
	standing.frames[6]=standing00;
	standing.frames[7]=standing00;
	standing.frames[8]=standing00;
	standing.frames[9]=standing00;
	standing.frames[10]=standing00;
	standing.frames[11]=standing00;
	standing.frames[12]=standing00;
	standing.frames[13]=standing00;
	standing.frames[14]=standing00;
	standing.frames[15]=standing00;
	standing.frames[16]=standing00;
	standing.frames[17]=standing00;
	standing.frames[18]=standing01;
	standing.frames[19]=standing02;
	standing.loopframe=0;
	standing.lastframe=19;
    standing.current_frame=0;

    sstanding.frames[0]=sstanding00;
	sstanding.frames[1]=sstanding00;
	sstanding.frames[2]=sstanding00;
	sstanding.frames[3]=sstanding00;
	sstanding.frames[4]=sstanding00;
	sstanding.frames[5]=sstanding00;
	sstanding.frames[6]=sstanding00;
	sstanding.frames[7]=sstanding00;
	sstanding.frames[8]=sstanding00;
	sstanding.frames[9]=sstanding00;
	sstanding.frames[10]=standing00;
	sstanding.frames[11]=sstanding00;
	sstanding.frames[12]=sstanding00;
	sstanding.frames[13]=sstanding00;
	sstanding.frames[14]=sstanding00;
	sstanding.frames[15]=sstanding00;
	sstanding.frames[16]=sstanding00;
	sstanding.frames[17]=sstanding00;
	sstanding.frames[18]=sstanding00;
	sstanding.frames[19]=sstanding00;

    bullet0.frames[0]=chargeL000;
	bullet0.frames[1]=chargeL001;
	bullet0.frames[2]=chargeL000;
	bullet0.frames[3]=chargeL001;
	bullet0.loopframe=0;
	bullet0.lastframe=3;
    bullet0.current_frame=0;

    bullet1.frames[0]=chargeL100;
	bullet1.frames[1]=chargeL101;
	bullet1.frames[2]=chargeL100;
	bullet1.frames[3]=chargeL101;
	bullet1.loopframe=0;
	bullet1.lastframe=3;
    bullet1.current_frame=0;

    bullet2.frames[0]=chargeL200;
	bullet2.frames[1]=chargeL201;
	bullet2.frames[2]=chargeL200;
	bullet2.frames[3]=chargeL201;
	bullet2.loopframe=0;
	bullet2.lastframe=3;
    bullet2.current_frame=0;
}

//////////////////////////////////////////////////////////////////////////
///////////////  update_animation() //////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void update_animation()
{
	int i;
	if ( player.doing==HURTING )
	{
		if ( hurting.current_frame>=hurting.lastframe )
			hurting.current_frame=hurting.loopframe;
		else if ( player.animpos>=player.animspeed )
			{ hurting.current_frame++; player.animpos=0; }
		else player.animpos++;
	}
	if ( player.doing!=HURTING)
	hurting.current_frame=0;

	if ( player.doing==WALKING )
	{
		if ( walking.current_frame>=walking.lastframe )
			walking.current_frame=walking.loopframe;
		else if ( player.animpos>=player.animspeed )
			{ walking.current_frame++; player.animpos=0; }
		else player.animpos++;

	}
	if ( player.doing!=WALKING)
		walking.current_frame=0;

	if ( player.doing==JUMPING )
	{
		if ( jumping.current_frame>=jumping.lastframe )
			jumping.current_frame=jumping.loopframe;
		else if ( player.animpos>=player.animspeed )
		    { jumping.current_frame++; player.animpos=0; }
		else player.animpos++;

	}
	if ( player.doing!=JUMPING)
		jumping.current_frame=0;

	if ( player.doing==FALLING )
	{
		if ( falling.current_frame>=falling.lastframe )
			falling.current_frame=falling.loopframe;
		else if ( player.animpos>=player.animspeed )
			{ falling.current_frame++; player.animpos=0; }
		else player.animpos++;

	}
	if ( player.doing!=FALLING)
		falling.current_frame=0;

	if ( player.doing==STANDING )
	{
		if ( standing.current_frame>=standing.lastframe )
			standing.current_frame=standing.loopframe;
		else if ( player.animpos>=player.animspeed )
			{ standing.current_frame++; player.animpos=0; }
		else player.animpos++;

	}
	if ( player.doing!=STANDING)
		standing.current_frame=0;

	if ( player.doing==SLIDING )
	{
		if ( sliding.current_frame>=sliding.lastframe )
			sliding.current_frame=sliding.loopframe;
		else if ( player.animpos>=player.animspeed )
			{ sliding.current_frame++; player.animpos=0; }
		else player.animpos++;

	}
	if ( player.doing!=SLIDING)
		sliding.current_frame=0;

	if ( player.doing==DASHING )
	{
		if ( dashing.current_frame>=dashing.lastframe )
			dashing.current_frame=dashing.loopframe;
		else if ( player.animpos>=player.animspeed )
			{ dashing.current_frame++; player.animpos=0; }
		else player.animpos++;

	}
	if ( player.doing!=DASHING)
		dashing.current_frame=0;

	if ( player.doing==KNEELING )
	{
		if ( kneeling.current_frame>=kneeling.lastframe )
			kneeling.current_frame=kneeling.loopframe;
		else if ( player.animpos>=player.animspeed )
			{ kneeling.current_frame++; player.animpos=0; }
		else player.animpos++;

	}
	if ( player.doing!=KNEELING)
		kneeling.current_frame=0;

	// Bullet animations..

	for ( i=0;i<MAXBULLETS;i++ )
    {
		if ( player.bullets[i].bullet.value1==TRUE )
		{
			if ( player.bullets[i].current_frame>=bullet0.lastframe )
				player.bullets[i].current_frame=bullet0.loopframe;
			else if ( player.bullets[i].animpos>=player.bullets[i].animspeed )
		    	{ player.bullets[i].current_frame++; player.bullets[i].animpos=0; }
     		else player.bullets[i].animpos++;
		}
		if ( player.bullets[i].bullet.value1==TRUE )
		{
			if ( player.bullets[i].current_frame>=bullet1.lastframe )
				player.bullets[i].current_frame=bullet1.loopframe;
			else if ( player.bullets[i].animpos>=player.bullets[i].animspeed )
		    	{ player.bullets[i].current_frame++; player.bullets[i].animpos=0; }
		   	else player.bullets[i].animpos++;
		}
		if ( player.bullets[i].bullet.value1==TRUE )
		{
			if ( player.bullets[i].current_frame>=bullet0.lastframe )
				player.bullets[i].current_frame=bullet2.loopframe;
			else if ( player.bullets[i].animpos>=player.bullets[i].animspeed )
		    	{ player.bullets[i].current_frame++; player.bullets[i].animpos=0; }
		   	else player.bullets[i].animpos++;
		}



	}

    /*
	for ( i=0;i<MAXBULLETS;i++ )
    {

	 if ( bullet0.current_frame>=bullet0.lastframe )
 			bullet0.current_frame=bullet0.loopframe;
     else if ( player.bullets[i].animpos>=player.bullets[i].animspeed )
		{ bullet0.current_frame++; player.bullets[i].animpos=0; }
     //else player.bullets[i].animpos++;

     if ( bullet1.current_frame>=bullet1.lastframe )
				bullet1.current_frame=bullet1.loopframe;
	 else if ( player.bullets[i].animpos>=player.bullets[i].animspeed )
		{ bullet1.current_frame++; player.bullets[i].animpos=0; }
	 //else player.bullets[i].animpos++;

	 if ( bullet2.current_frame>=bullet2.lastframe )
					bullet2.current_frame=bullet2.loopframe;
	 else if ( player.bullets[i].animpos>=player.bullets[i].animspeed )
		{ bullet2.current_frame++; player.bullets[i].animpos=0; }
	 //else player.bullets[i].animpos++;
    }*/

}

//////////////////////////////////////////////////////////////////////////
///////////////  init_engine() ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

int init_engine()
{
    fps2d_init();
    init_maps();
    fps2d_tiles( 16, 16 );
    setup_player();
    megamanx=load_datafile( "./data/megamanx.dat" );
    build_animations();
    init_enemies();

    return 0;
}

//////////////////////////////////////////////////////////////////////////
///////////////  gameloop() //////////////////////////////////////////////

void gameloop()
{
	exit_level=FALSE;
	play_mod( music_track, TRUE );
	while( exit_level == FALSE )
	{

		while (aspeed>gspeed)
		{
		  	getinput();
		  	gametick();
		  	doenemyai();
		  	update_animation();
		  	set_camera();

		  	gspeed++;
		}
		clear( buffer );
		fps2d_draw_map( current_map, buffer, 0 );
		draw_player();
		draw_enemies();
	    //draw_dustclouds();  ( goes in draw_player() )
		//draw_bullets();
		draw_lifebar( 5, 5 );

		/*text_mode(-1);
		textprintf( buffer, (FONT*)sprites[mmxfont].dat, 250, 5, -1, "FPS: %d", last_fps );
        textprintf( buffer, (FONT*)sprites[mmxfont].dat, 5, 5, -1, "enemy0x: %d", enemies[0].body[0].x );
        textprintf( buffer, (FONT*)sprites[mmxfont].dat, 5, 13, -1, "enemy0y: %d", enemies[0].body[0].y );
        textprintf( buffer, (FONT*)sprites[mmxfont].dat, 5, 21, -1, "enemy0dir: %d", enemies[0].body[0].dir );
        textprintf( buffer, (FONT*)sprites[mmxfont].dat, 5, 29, -1, "enemy0estuck: %d", estuck(enemies[0]) );
        textprintf( buffer, (FONT*)sprites[mmxfont].dat, 5, 37, -1, "enemy0einair: %d", einair(enemies[0]) );
        textprintf( buffer, (FONT*)sprites[mmxfont].dat, 5, 45, -1, "enemy0doing: %d", enemies[0].doing );
        textprintf( buffer, (FONT*)sprites[mmxfont].dat, 5, 53, -1, "enemy0life: %d", enemies[0].life );
        */
		blit( buffer, screen, 0, 0, 0, 0, 320, 240 );
	    frame_counter++;
	}

}

//////////////////////////////////////////////////////////////////////////
///////////////  startmap() //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

int start_map( MAP &m )
{
	init_engine();
	if ( fps2d_load_props(1, m.propfile) == NULL )
	{
		sprintf( temps, "Unable to find file: %s", m.propfile );
	    debuglog( temps );
	    exit(0);
	}
    if ( File_is_There( m.mapfile ) == 0 )
    {
		sprintf( temps, "Unable to find file: %s", m.mapfile );
	    debuglog( temps );
	    exit(0);
	}
    else current_map=fps2d_load_maps( m.mapfile );

    if ( fps2d_load_textures( m.tilefile ) != 0 )
    {
		sprintf( temps, "Unable to find file: %s", m.tilefile );
        debuglog( temps );
        exit(0);
	}

    if( File_is_There( m.musicfile ) == 0 )
    {
		sprintf( temps, "Unable to find file: %s", m.musicfile );
		debuglog( temps );
		exit(0);
	}
	else music_track=load_mod( m.musicfile );

    gameloop();

    return 0;
}

//////////////////////////////////////////////////////////////////////////
///////////////  init_maps() /////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void init_maps()
{
	maps[0].tilefile="./data/level0.dat";
 	maps[0].propfile="./data/level0.op";
 	maps[0].mapfile="./data/level0.map";
	maps[0].title="./data/Test Level";
	maps[0].musicfile="./data/test.xm";
}

//////////////////////////////////////////////////////////////////////////
///////////////  draw_screen() ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void draw_screen()
{
	fps2d_draw_map( current_map, buffer, 0 );
	vsync();
	blit( buffer, screen, 0, 0, 0, 0, 320, 240 );
}

//////////////////////////////////////////////////////////////////////////
///////////////  getinput() //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void getinput()
{
	int i,j;

	if ( key[KEY_O] )
	 player.chp--;
	if ( key[KEY_P] )
	 player.chp++;


	if ( key[KEY_ESC] )
	{
		exit_level=TRUE;
	}

    if ( key[KEY_LEFT] && player.doing!=HURTING )
    {
   		player.facing=LEFT;
  		if ( inair()==0 && !key[KEY_DOWN] )
  		{
   			player.doing=WALKING;
  		}
 	}

	if ( key[KEY_RIGHT] && player.doing!=HURTING )
	{
		player.facing=RIGHT;
 		if ( inair()==0 && !key[KEY_DOWN] )
 		{
 			player.doing=WALKING;
 		}
 	}

 if ( key[KEY_RSHIFT] && player.doing!=HURTING )
 {
  if ( jumppressed==FALSE )
   {
    if ( player.doing==WALKING || player.doing==STANDING || player.doing==DASHING)
     {
      player.doing=JUMPING;
      play_sample( (SAMPLE*)sounddat[jump00].dat, 255, 128, 1000, FALSE);
     }
    if ( player.doing==SLIDING )
     {
      player.doing=JUMPING;
      play_sample( (SAMPLE*)sounddat[jump00].dat, 255, 128, 1000, FALSE);
      if ( key[KEY_RCONTROL] )
       player.dashjump=TRUE;
     }
    jumppressed=TRUE;
   }

 }

 if( !key[KEY_LEFT] && !key[KEY_RIGHT] && player.doing!=JUMPING && inair()==0 && player.doing!=HURTING)
 {
  player.doing=STANDING;
 }

 if ( key[KEY_ENTER] && player.doing!=HURTING )
  {
   if ( shootpressed==FALSE )
   {
    // are all shots fired?
    for ( j=0,i=0;i<MAXBULLETS;i++ )
    {
     if ( player.bullets[i].bullet.value1==FALSE && j==0)
      {
       player.bullets[i].bullet.x=player.body.x;
       player.bullets[i].bullet.y=player.body.y-8;
       player.bullets[i].level=0;
       if ( player.doing==SLIDING )
        player.bullets[i].bullet.dir=reversedir(player.facing);
       else
        player.bullets[i].bullet.dir=player.facing;
       player.bullets[i].bullet.value1=TRUE; j=1;
       player.shooting=TRUE;
       play_sample( (SAMPLE*)sounddat[shot00].dat, 255, 128, 1000, FALSE);
      }
    }
   }
   shootpressed=TRUE;
  }


 if ( key[KEY_DOWN] )
 {
	 player.doing=KNEELING;
 }

 if ( key[KEY_RCONTROL] && player.doing!=HURTING )
 {
 if ( dashpressed==FALSE && !key[KEY_RSHIFT] )
  { player.doing=DASHING;
    if ( dashsoundplaying==FALSE )
      play_sample( (SAMPLE*)sounddat[dash00].dat, 255, 128, 1000, FALSE);
    dashsoundplaying=TRUE;}
 else if ( dashpressed==FALSE && key[KEY_RSHIFT] && inair()==0)
  { player.dashjump=TRUE; }
 }

}

//////////////////////////////////////////////////////////////////////////
///////////////  reversedir() ////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

char reversedir( char dir )
{
	if ( dir==LEFT )
		return RIGHT;

	return LEFT;
}

//////////////////////////////////////////////////////////////////////////
///////////////  set_camera() ////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void set_camera()
{
  if ( player.body.x<=160 && player.body.x>=0
       && player.body.y>=current_map->map->layer->height*16-120
       && player.body.y<=current_map->map->layer->height*16 )
   fps2d_center_camera(current_map, 160, current_map->map->layer->height*16-120, 0);

  else if ( player.body.x>=160 && player.body.x<=current_map->map->layer->width*16-160
            && player.body.y>=current_map->map->layer->height*16-120
            && player.body.y<=current_map->map->layer->height*16 )
   fps2d_center_camera(current_map, player.body.x, current_map->map->layer->height*16-120, 0);

  else if ( player.body.x>=current_map->map->layer->width*16-160
            && player.body.x<=current_map->map->layer->width*16
            && player.body.y>=current_map->map->layer->height*16-120
            && player.body.y<=current_map->map->layer->height*16 )
   fps2d_center_camera(current_map, current_map->map->layer->width*16-160, current_map->map->layer->height*32-120, 0);

  else if ( player.body.x<=160 && player.body.x>=0 && player.body.y>=0 && player.body.y<=120 )
   fps2d_center_camera(current_map, 160, 120, 0);

  else if ( player.body.x>=160 && player.body.x<=current_map->map->layer->width*16-160
            && player.body.y<=120 && player.body.y>= 0)
   fps2d_center_camera(current_map, player.body.x, 120, 0);

  else if ( player.body.x>=current_map->map->layer->width*16-160
            && player.body.x<=current_map->map->layer->width*16
            && player.body.y>=0 && player.body.y<=120 )
   fps2d_center_camera(current_map, current_map->map->layer->width*16-160, 120, 0);

  else if ( player.body.x<=160 && player.body.x>=0
            && player.body.y>=120 && player.body.y<=current_map->map->layer->height*16-120 )
   fps2d_center_camera(current_map, 160, player.body.y, 0);

  else if ( player.body.x>=current_map->map->layer->width*16-160 && player.body.x<=current_map->map->layer->width*32
            && player.body.y>=120 && player.body.y<=current_map->map->layer->height*16-120 )
   fps2d_center_camera(current_map, current_map->map->layer->width*16-160, player.body.y, 0);

  else fps2d_center_camera(current_map, player.body.x, player.body.y, 0);
}

char stuck()
{
	if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
		return 1;

 return 0;
}

char estuck( ENEMYDAT &e )
{
	if ( fps2d_map_xy_collision(current_map, e.body[0], 1, 0, 0) == 1 )
		return 1;

 return 0;
}

//////////////////////////////////////////////////////////////////////////
///////////////  touchingwall() //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

char touchingwall()
{
	player.body.x--;
 	if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
 	{
		player.body.x++;
		return 1;
	}
 	player.body.x++;

 	player.body.x++;
 	if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
 	{
		player.body.x--;
		return 2;
	}
 	player.body.x--;

 	return 0;
}

//////////////////////////////////////////////////////////////////////////
///////////////  inair() /////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

char inair()
{
	player.body.y++;
	if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
	{
		player.body.y--;
		return 0;
	}
 	player.body.y--;

 	return 1;
}

char einair( ENEMYDAT &e )
{
	e.body[0].y++;
	if ( fps2d_map_xy_collision(current_map, e.body[0], 1, 0, 0) == 1 )
	{
		e.body[0].y--;
		return 0;
	}
 	e.body[0].y--;

 	return 1;
}

//////////////////////////////////////////////////////////////////////////
///////////////  setup_player() //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


void setup_player()
{
	int i;

	player.body.width=Tile_Width;
	player.body.height=Tile_Height;
	player.head.width=Tile_Width;
	player.head.height=Tile_Height;
	player.body.x=Tile_Width;
	player.body.y=Tile_Height;
	player.facing=RIGHT;
	player.speed=2;
	player.doing=STANDING;
    player.jumpspeed=3;
    player.fallspeed=4;
    player.maxjumpheight=(16*4)/player.jumpspeed;
    player.dashdistance=20;
    player.animspeed=2;
    player.animpos=0;

    player.hurtable=TRUE;
    player.hurtdur=30;

	player.hp=20;
	player.chp=20;
	player.ep=20;
	player.cep=20;
	player.ap=20;
	player.cap=20;

	player.chargelevel0=50;
    player.chargelevel1=150;
	player.chargelevel2=250;
	player.chargelevel0c=makecol(0,255,255);
  	player.chargelevel1c=makecol(0,255,0);

	player.shooting=FALSE;
	player.shootingpos=0;
	player.shootingdur=20;

	player.chargeflash=FALSE;
	player.chargeflashpos=0;
	player.chargeflashdur=2;
	player.alreadycharging=FALSE;

	player.hurtflash=FALSE;
	player.hurtflashpos=0;
	player.hurtflashdur=2;

	player.hurt_flash=FALSE;
	player.hurt_flash_pos=0;
	player.hurt_flash_dur=100;

    for ( i=0; i<MAXBULLETS; i++ )
	{
		player.bullets[i].bullet.width=Tile_Width;
 		player.bullets[i].bullet.height=Tile_Height;
		player.bullets[i].bullet.value1=FALSE;
		player.bullets[i].speed=5;
		player.bullets[i].level=0;
		player.bullets[i].animpos=0;
		player.bullets[i].animspeed=2;
		player.bullets[i].bullet.x=player.bullets[i].bullet.y=0;
	}

	for ( i=0; i<MAXDUSTCLOUDS; i++ )
	{
		player.dustclouds[i].dustcloud.width=Tile_Width;
		player.dustclouds[i].dustcloud.height=Tile_Height;
		player.dustclouds[i].dustcloud.value1=FALSE;
		player.dustclouds[i].dustcloud.x=player.dustclouds[i].dustcloud.y=0;
 	}
}

//////////////////////////////////////////////////////////////////////////
///////////////  gametick() //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void gametick()
{
	//player hurting?
	if ( player.doing==HURTING )
	{
		if ( player.hurtpos>=player.hurtdur )
		{
			player.doing=STANDING;
			player.hurtpos=0;

			player.hurtflash=TRUE;
		}
		else
		{
			/*if ( player.facing==RIGHT )
				player.body.x--;
			if ( player.facing==LEFT )
				player.body.y++;*/
			player.hurtpos++;
		}
	}

	// player dead?
	if ( player.chp<=0 )
		do_playerdeath();

	if ( player.chargelevel>=player.chargelevel0 &&
	     player.chargelevel<=player.chargelevel1 )
	  player.chargecolor=player.chargelevel0c;
	else if ( player.chargelevel>=player.chargelevel1 &&
	     player.chargelevel<=player.chargelevel2 )
	  player.chargecolor=player.chargelevel1c;

	int i,j;
  //if bullets exist.. move them.
 for ( i=0;i<MAXBULLETS;i++ )
  {
   if ( player.bullets[i].bullet.value1==TRUE )
   {
    if ( player.bullets[i].bullet.dir==LEFT )
     player.bullets[i].bullet.x-=player.bullets[i].speed;
    else if (player.bullets[i].bullet.dir==RIGHT )
     player.bullets[i].bullet.x+=player.bullets[i].speed;
   }
  }

 // bullet gone off right side?
 for ( i=0;i<MAXBULLETS;i++ )
  {
   if ( player.bullets[i].bullet.x>=current_map->map[0].cam_x+320 )
    player.bullets[i].bullet.value1=FALSE;
  }
 // bullet gone off left side?
 for ( i=0;i<MAXBULLETS;i++ )
  {
   if ( player.bullets[i].bullet.x<=current_map->map[0].cam_x )
    player.bullets[i].bullet.value1=FALSE;
  }

 // Bullet hit wall?
 for ( i=0;i<MAXBULLETS;i++ )
  {
   if ( fps2d_map_xy_collision(current_map, player.bullets[i].bullet, 1, 0, 0) == 1 )
        player.bullets[i].bullet.value1=FALSE;
  }


 // If player is in air.. make player fall.
 if( player.doing!=JUMPING && inair()==1 )
  {
   if ( player.doing==SLIDING )
   { player.body.y+=player.fallspeed/2; }
   else {player.body.y+=player.fallspeed; player.doing=FALLING;}
  }


 // If player is trying to move in air...
 if ( player.doing!=STANDING && player.doing!=WALKING && player.doing!=DASHING && player.doing!=KNEELING && player.doing!=HURTING)
  {
   if ( player.facing==RIGHT )
     {
     if ( key[KEY_RIGHT] )
      {
       if ( player.dashjump==TRUE )
        {
         player.body.x+=player.speed*2;
         if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
          player.body.x-=player.speed*2;
        }
       else
        {
         player.body.x+=player.speed;
         if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
          player.body.x-=player.speed;
        }
      }
     }
     if ( player.facing==LEFT )
      {
       if ( key[KEY_LEFT] )
        {
         if ( player.dashjump==TRUE )
          {
           player.body.x-=player.speed*2;
           if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
            player.body.x+=player.speed*2;
          }
         else
          {
           player.body.x-=player.speed;
           if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
            player.body.x+=player.speed;
          }
        }
      }
  }

 // if player is walking..
 if ( player.doing==WALKING )
  {
   if ( player.facing==RIGHT )
    {
     player.body.x+=player.speed;
     if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
      player.body.x-=player.speed;
    }
   if ( player.facing==LEFT )
    {
     player.body.x-=player.speed;
     if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
      player.body.x+=player.speed;
    }
  }

  // if player is dashing..
 if ( player.doing==DASHING )
  {
   if ( player.facing==RIGHT && player.dashpos<=player.dashdistance && key[KEY_RCONTROL])
    {
     player.body.x+=player.speed*2;
     if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
      player.body.x-=player.speed*2;
     player.dashpos++;
    }
   else if ( player.facing==LEFT && player.dashpos<=player.dashdistance && key[KEY_RCONTROL])
    {
     player.body.x-=player.speed*2;
     if ( fps2d_map_xy_collision(current_map, player.body, 1, 0, 0) == 1 )
      player.body.x+=player.speed*2;
     player.dashpos++;
    }
   else
   { player.dashpos=0; player.doing=STANDING; dashpressed=TRUE;}
  }

 // if player is jumping..
 if ( player.doing==JUMPING )
  {
   if ( player.jumpheight>player.maxjumpheight )
    { player.doing=FALLING; player.jumpheight=0; }
   else if ( fps2d_map_xy_collision(current_map, player.head, 1, 0, 0) == 1 )
    { player.doing=FALLING; player.jumpheight=0; }
   else if ( !key[KEY_RSHIFT] )
    { player.doing=FALLING; player.jumpheight=0; }
   else { player.body.y-=player.jumpspeed; player.jumpheight++; }
  }

 // player sliding..
 if ( player.doing==FALLING )
 {
  player.dashpos=0; dashpressed=TRUE;
  if (inair()==0)
   { player.doing=FALLING; }
  else if ( touchingwall()==2 && player.facing==RIGHT && key[KEY_RIGHT])
  {
   player.doing=SLIDING;
  }
  else if ( touchingwall()==1 && player.facing==LEFT && key[KEY_LEFT])
  {
   player.doing=SLIDING;
  }
 }

 // unstick..
 if ( stuck()==1)
  {
   if( player.doing==JUMPING )
     while( stuck()==1 ) player.body.y++;
   else { while(stuck()==1) player.body.y--; play_sample( (SAMPLE*)sounddat[land00].dat, 255, 128, 1000, FALSE);}
  }

  if ( !key[KEY_RSHIFT] )
   jumppressed=FALSE;
  if ( !key[KEY_RCONTROL] )
   { dashpressed=FALSE; dashsoundplaying=FALSE; }
  if ( !key[KEY_ENTER] )
   {
    stop_sample((SAMPLE*)sounddat[charge00].dat);
    stop_sample((SAMPLE*)sounddat[chargestart00].dat);

    if ( player.chargelevel>=player.chargelevel0 && player.chargelevel<player.chargelevel1)
    {

    // are all shots fired?
	    for ( j=0,i=0;i<MAXBULLETS;i++ )
	    {
	     if ( player.bullets[i].bullet.value1==FALSE && j==0)
	      {
	       player.bullets[i].bullet.x=player.body.x;
	       player.bullets[i].bullet.y=player.body.y-8;
	       player.bullets[i].level=1;
	       if ( player.doing==SLIDING )
	        player.bullets[i].bullet.dir=reversedir(player.facing);
	       else
	        player.bullets[i].bullet.dir=player.facing;
	       player.bullets[i].bullet.value1=TRUE; j=1;

		   player.shooting=TRUE;
	       play_sample( (SAMPLE*)sounddat[shot01].dat, 255, 128, 1000, FALSE);
	      }
       }


    }
    else if ( player.chargelevel>=player.chargelevel1 && player.chargelevel<=player.chargelevel2)
    {

    // are all shots fired?
	    for ( j=0,i=0;i<MAXBULLETS;i++ )
	    {
	     if ( player.bullets[i].bullet.value1==FALSE && j==0)
	      {
	       player.bullets[i].bullet.x=player.body.x;
	       player.bullets[i].bullet.y=player.body.y-8;
	       player.bullets[i].level=2;
	       if ( player.doing==SLIDING )
	        player.bullets[i].bullet.dir=reversedir(player.facing);
	       else
	        player.bullets[i].bullet.dir=player.facing;
	       player.bullets[i].bullet.value1=TRUE; j=1;

		   player.shooting=TRUE;
	       play_sample( (SAMPLE*)sounddat[shot02].dat, 255, 128, 1000, FALSE);
	      }
       }


    }
    shootpressed=FALSE; player.chargelevel=0;

   }
  if ( key[KEY_ENTER] )
   {
    if ( player.chargelevel<MAXCHARGELEVEL )
     player.chargelevel++;
   }

  // reset dashjump
 if ( player.doing!=JUMPING && player.doing!=FALLING )
 {
  player.dashjump=FALSE;
 }
 /*
 // dustcloud animation..
 for(i=0;i<MAXDUSTCLOUDS;i++)
 {
  if ( dustcloud[i].value1==TRUE )
  {
   if ( dustcloud[i].value2==maxdustframes )
   {
    dustcloud[i].value1=FALSE;
    dustcloud[i].value2=0;
   }
   if ( dustanimpos>=dustspeed )
    { dustcloud[i].value2++; dustanimpos=0; }
   else dustanimpos++;
  }
 }
 */

 // weapon charge sounds..
 if ( player.chargelevel>=player.chargelevel0 && player.chargelevel<=player.chargelevel1 && player.alreadycharging==FALSE )
 {
  player.alreadycharging=TRUE;
  play_sample( (SAMPLE*)sounddat[chargestart00].dat, 255, 128, 1000, FALSE);
  player.chargeflash=TRUE;
 }

 if ( player.chargelevel>=player.chargelevel1 && player.chargelevel<=player.chargelevel2 )
 {
  play_sample( (SAMPLE*)sounddat[charge00].dat, 255, 128, 1000, TRUE);
 }

 if( !key[KEY_RCONTROL] )
  player.dashpos=0;
 /*
 // make dust clouds..
 if ( playercurrdoing==DASHING )
 {
  for( i=0,j=0,k=0; i<MAXDUSTCLOUDS; i++ )
   {
    if ( dustcloud[i].value1==FALSE && dustspawn>=2)
    {
     dustcloud[i].value1=TRUE;
     dustcloud[i].x=player.x-j*k;
     dustcloud[i].y=player.y+16;
     dustspawn=0;
     j+=32;
     k++;
    }
    dustcloud[i].y=dustcloud[i].y-=2;
   }
   dustspawn++;
 } */

 if ( player.hurt_flash_pos>=player.hurt_flash_dur )
 	{
		player.hurt_flash=FALSE;
		player.hurt_flash_pos=0;

		player.hurtable=TRUE;
	}
 else player.hurt_flash_pos++;

 if ( player.shooting==TRUE )
 {
 	if ( player.shootingpos >= player.shootingdur )
 	  { player.shooting=FALSE; player.shootingpos=0; }
 	else player.shootingpos++;
 }

 if ( player.hurtflash==TRUE )
 {
  	if ( player.hurtflashpos >= player.hurtflashdur )
  	  { player.hurtflash=FALSE; player.hurtflashpos=0; }
  	else player.hurtflashpos++;
 }
 if ( player.hurtflash==FALSE )
 {
   	if ( player.hurtflashpos >= player.hurtflashdur && player.hurt_flash==TRUE)
   	  { player.hurtflash=TRUE; player.hurtflashpos=0; }
   	else player.hurtflashpos++;
 }

 if ( player.chargeflash==TRUE )
 {
  	if ( player.chargeflashpos >= player.chargeflashdur )
  	  { player.chargeflash=FALSE; player.chargeflashpos=0; player.alreadycharging=FALSE;}
  	else player.chargeflashpos++;
 }
 if ( player.chargeflash==FALSE )
 {
   	if ( player.chargeflashpos >= player.chargeflashdur && player.chargelevel>=player.chargelevel0)
   	  { player.chargeflash=TRUE; player.chargeflashpos=0; }
   	else player.chargeflashpos++;
 }

 // Color flash, during charge..
 if ( player.chargeflash==TRUE )
 {
	 if ( !key[KEY_ENTER] )
	  player.chargeflash=FALSE;
 }


 player.head.x=player.body.x;
 player.head.y=player.body.y-16;
}

//////////////////////////////////////////////////////////////////////////
///////////////  draw_player() ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void draw_megaman(ANIM* strand, int x, int y, int left) {
	BITMAP* bmp = (BITMAP*)megamanx[strand->frames[strand->current_frame]].dat;
	if (left) {
		draw_sprite_h_flip(buffer, bmp, x, y);
	} else {
		draw_sprite(buffer, bmp, x, y);
	}
}

void draw_charge(ANIM* strand, int x, int y, int left) {
	if (player.chargeflash == FALSE) {
		return;
	}
	BITMAP* bmp = (BITMAP*)megamanx[strand->frames[strand->current_frame]].dat;
	if (left) {
		draw_trans_sprite_h_flip(buffer, sil(bmp, player.chargecolor), x, y);
	} else {
		draw_trans_sprite(buffer, sil(bmp, player.chargecolor), x, y);
	}
}

void draw_blankman() {
	int x = player.body.x - current_map->map[0].cam_x;
	int y = player.body.y - current_map->map[0].cam_y;
	rectfill(buffer, x, y, x + Tile_Width, y + Tile_Height, makecol(255, 255, 255));
	x = player.head.x - current_map->map[0].cam_x;
	y = player.head.y - current_map->map[0].cam_y;
	rectfill(buffer, x, y, x + Tile_Width, y + Tile_Height, makecol(255, 0, 0));
}

void draw_player() {
	if (player.hurtflash == FALSE) {
		ANIM* strand;
		int hurt = 0;
		int blankman = 0;
		switch (player.doing) {
			//case HURTING: strand = hurting; hurt = TRUE; break;
			case WALKING: strand = player.shooting ? &swalking : &walking; break;
			case JUMPING: strand = player.shooting ? &sjumping : &jumping; break;
			case FALLING: strand = player.shooting ? &sfalling : &falling; break;
			case STANDING:strand = player.shooting ? &sstanding: &standing;break;
			case SLIDING: strand = player.shooting ? &ssliding : &sliding; break;
			case DASHING: strand = player.shooting ? &sdashing : &dashing; break;
			case KNEELING:strand = player.shooting ? &skneeling: &kneeling;break;
			default: blankman = 1;
		}
		if (blankman) {
			draw_blankman();
		} else {
			int x = player.body.x - current_map->map[0].cam_x - 8;
			int y = player.body.y - current_map->map[0].cam_y - 16;
			draw_megaman(strand, x, y, player.facing == LEFT);
			if (!hurt) {
				draw_charge(strand, x, y, player.facing == LEFT);
			}
		}
	}
// draw player related objects.
	draw_bullets();
}

void draw_bullets()
{
	int i;
 for ( i=0;i<MAXBULLETS;i++ )
  {
   if ( player.bullets[i].bullet.value1==TRUE && player.bullets[i].level==0 )
   {
   		if ( player.bullets[i].bullet.dir==LEFT )
   			draw_sprite_h_flip( buffer, (BITMAP*)sprites[bullet0.frames[player.bullets[i].current_frame]].dat,
                                player.bullets[i].bullet.x - current_map->map[0].cam_x,
                                player.bullets[i].bullet.y - current_map->map[0].cam_y );
        else
   		    draw_sprite( buffer, (BITMAP*)sprites[bullet0.frames[player.bullets[i].current_frame]].dat,
                                player.bullets[i].bullet.x - current_map->map[0].cam_x,
	                            player.bullets[i].bullet.y - current_map->map[0].cam_y );
   }
   if ( player.bullets[i].bullet.value1==TRUE && player.bullets[i].level==1 )
      {
      		if ( player.bullets[i].bullet.dir==LEFT )
      			draw_sprite_h_flip( buffer, (BITMAP*)sprites[bullet1.frames[player.bullets[i].current_frame]].dat,
                                   player.bullets[i].bullet.x - current_map->map[0].cam_x,
                                   player.bullets[i].bullet.y - current_map->map[0].cam_y );
           else
      		    draw_sprite( buffer, (BITMAP*)sprites[bullet1.frames[player.bullets[i].current_frame]].dat,
                                   player.bullets[i].bullet.x - current_map->map[0].cam_x,
   	                            player.bullets[i].bullet.y - current_map->map[0].cam_y );
   }
   if ( player.bullets[i].bullet.value1==TRUE && player.bullets[i].level==2 )
      {
      		if ( player.bullets[i].bullet.dir==LEFT )
      			draw_sprite_h_flip( buffer, (BITMAP*)sprites[bullet2.frames[player.bullets[i].current_frame]].dat,
                                   player.bullets[i].bullet.x - current_map->map[0].cam_x,
                                   player.bullets[i].bullet.y - current_map->map[0].cam_y );
           else
      		    draw_sprite( buffer, (BITMAP*)sprites[bullet2.frames[player.bullets[i].current_frame]].dat,
                                   player.bullets[i].bullet.x - current_map->map[0].cam_x,
   	                            player.bullets[i].bullet.y - current_map->map[0].cam_y );
   }


  }
}

BITMAP* sil( BITMAP* bmp, int c )
{
	int x,y,color,r,g,b;
	BITMAP* b2=create_bitmap( bmp->w, bmp->h );

	for ( x=0 ; x < bmp->w ; x++ )
	{
		for ( y=0 ; y < bmp->h ; y++ )
		{
			color=getpixel( bmp, x, y );
			r=getr(color);
			g=getg(color);
			b=getb(color);

			if ( color!=makecol(255, 0, 255) )
			 putpixel( b2, x, y, c );
			else
			 putpixel( b2, x, y, makecol(r,g,b) );
		}

	}
	return b2;
}

void do_playerdeath()
{
	// death animation.

	start_map( maps[0] );
}

void draw_trans_sprite_h_flip( BITMAP* dest, BITMAP* bmp, int x, int y )
{
	BITMAP* b=create_bitmap( bmp->w, bmp->h );
	clear_to_color( b, makecol( 255, 0, 255 ));
	draw_sprite_h_flip( b, bmp, 0, 0 );

	draw_trans_sprite( dest, b, x, y );
}

void draw_lifebar( int x, int y )
{
	draw_sprite( buffer, (BITMAP*)sprites[lifebar00].dat, x, y );
	draw_lifemeter( player.hp, player.chp, x+17, y+3, 67 );
	textprintf( buffer, (FONT*)sprites[mmxfont].dat, x+89, y+4, -1, "%d", player.chp );
	textprintf( buffer, (FONT*)sprites[mmxfont].dat, x+89, y+16, -1, "%d", player.cep );

}

void draw_lifemeter( int max, int curr, int x, int y, int size )
{
	float p,q;
	p=(float)curr/(float)max;
	q=p*size;
	if ( p>1.00)
	{
		rectfill( buffer, x, y, x+(int)q, y+8, makecol(0,0,255) );
		textout( buffer, (FONT*)sprites[mmxfont].dat, "Cheater!!!!", x, y, -1 );
	}
	if ( p<=1.00 && p>0.50)
		rectfill( buffer, x, y, x+(int)q, y+8, makecol(0,255,0) );
	if ( p<=0.50 && p>0.25)
		rectfill( buffer, x, y, x+(int)q, y+8, makecol(255,255,0) );
	if ( p<=0.25 && p>0.00)
		rectfill( buffer, x, y, x+(int)q, y+8, makecol(255,0,0) );
}

void draw_enemies()
{
	int i;
	for ( i=0; i<=total_enemies; i++ )
	{
		if ( enemies[i].active == TRUE )
		{
			if ( enemies[i].body[0].dir==RIGHT )
				draw_sprite( buffer, (BITMAP*)sprites[spikeguy00].dat,
				    enemies[i].body[0].x - current_map->map[0].cam_x,
                    enemies[i].body[0].y - current_map->map[0].cam_y );
			else
				draw_sprite_h_flip( buffer, (BITMAP*)sprites[spikeguy00].dat,
				    enemies[i].body[0].x - current_map->map[0].cam_x,
                    enemies[i].body[0].y - current_map->map[0].cam_y );
		}
	}
}

void init_enemies()
{
	int i;
	total_enemies=10;

	enemies[0].body[0].x=enemies[0].body[0].y=50;
	enemies[0].body[0].dir=RIGHT;
	enemies[0].body[0].speed=1;
	enemies[0].body[0].width=16;
	enemies[0].body[0].height=200;
	enemies[0].doing=WALKING;
	enemies[0].life=10;
	enemies[0].active=TRUE;
	enemies[0].index=ROLLERGUY;
	enemies[0].cdmg=1;

	for ( i=1; i<=total_enemies; i++ )
	{
		enemies[i]=enemies[0];
		enemies[i].body[0].x=rand()%400+200;
	}

}

void doenemyai()
{
	int i;
	for ( i=0; i<=total_enemies; i++ )
	{
    	if ( enemies[i].index==ROLLERGUY )
  			rollerguyai(enemies[i]);
	}
}

void rollerguyai( ENEMYDAT &e )
{
	int i;
	if ( e.active==TRUE )
	{
		// in air? fall.
		if( einair(e)==1 )
		{ e.body[0].y++; e.doing=FALLING; }

		if ( e.body[0].dir==RIGHT )
		    {
		     e.body[0].x+=e.body[0].speed;
		    if ( fps2d_map_xy_collision(current_map, e.body[0], 1, 0, 0) == 1 )
		      { e.body[0].x-=e.body[0].speed; e.body[0].dir=LEFT; }
		    }
		if ( e.body[0].dir==LEFT )
		    {
		     e.body[0].x-=e.body[0].speed;
		    if ( fps2d_map_xy_collision(current_map, e.body[0], 1, 0, 0) == 1 )
		      { e.body[0].x+=e.body[0].speed; e.body[0].dir=RIGHT; }
		    }
		// hit by bullet?
		for ( i=0; i<MAXBULLETS; i++ )
		{
			if ( fps2d_object_collision( e.body[0], player.bullets[i].bullet ) == 1 && player.bullets[i].bullet.value1==TRUE)
		   	{
				player.bullets[i].bullet.value1=FALSE;
				e.life--;
				if ( e.life<=0 )
					e.active=FALSE;
			}
		}

		//hit by player?
		if ( fps2d_object_collision( e.body[0], player.body ) == 1 && player.hurtable==TRUE )
		{
			player.chp-=e.cdmg;
			player.doing=HURTING;
			player.hurtable=FALSE;
			player.hurt_flash=TRUE;
		}


    	// unstick..
   		//if ( estuck(e)==1)
    	while(einair(e)==0)
    	  { e.body[0].y--; e.doing=WALKING; }
	}

}
